class Policy(object):
    def __init__(self, eval:bool):
        self.eval = eval

    def choose_action(self, values=(), optimal=False):
        raise NotImplementedError

    def update(self, *args, **kwargs):
        pass

    def reset(self, *args, **kwargs):
        pass

    def set_eval(self, eval: bool):
        self.eval = eval



# -*- coding: utf-8 -*-

"""Top-level package for RLTG."""

__author__ = """Marco Favorito"""
__email__ = 'marco.favorito@gmail.com'
__version__ = '0.1.2'

import logging
logging.getLogger('rltg').addHandler(logging.NullHandler())
logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

logger = logging.getLogger('matplotlib')
# set WARNING for Matplotlib
logger.setLevel(logging.WARNING)





import logging
import os
import random
import shutil
import time

from rltg.agents.Agent import Agent
from rltg.trainers.Trainer import DEFAULT_DATA_DIR, Trainer
from rltg.utils.GoalEnvWrapper import GoalEnvWrapper
from rltg.utils.StatsManager import StatsManager
from rltg.utils.StoppingCondition import GoalPercentage

import pickle

from rltg.utils.misc import logger_from_verbosity


class GenericTrainer(Trainer):

    def __init__(self, env:GoalEnvWrapper, agent:Agent, n_episodes=1000,
                 data_dir=DEFAULT_DATA_DIR,
                 stop_conditions=(GoalPercentage(10, 1.0), ),):
        self.env = env
        self.n_episodes = n_episodes
        self.stop_conditions = list(stop_conditions)

        self.data_dir = data_dir
        self.agent_data_dir = data_dir + "/agent_data"

        self.agent = agent

        if not os.path.isdir(self.agent_data_dir):
            os.makedirs(self.agent_data_dir, exist_ok=True)

        self.cur_episode = 0
        self.stats = StatsManager(name="train_stats")
        self.optimal_stats = StatsManager(name="eval_stats")


    def main(self, eval:bool=False, render:bool=False, verbosity:int=1):
        logger = logger_from_verbosity(verbosity)

        agent = self.agent
        num_episodes = self.n_episodes
        stats = self.stats
        optimal_stats = self.optimal_stats

        if eval:
            stats.reset()
            optimal_stats.reset()
            self.cur_episode = 0

        for ep in range(self.cur_episode, num_episodes):

            if not eval:
                steps, total_reward, goal = self.train_loop(render=render)
                stats.update(steps, len(agent.brain.Q), total_reward, goal)
                summary = stats.print_summary(ep, steps, len(agent.brain.Q), total_reward, agent.brain.policy.epsilon.get(), goal)
                logger.info(summary)

            # try optimal run
            agent.set_eval(True)
            steps, total_reward, goal = self.train_loop(render=render)
            optimal_stats.update(steps, len(agent.brain.Q), total_reward, goal)
            optimal_summary = optimal_stats.print_summary(ep, steps, len(agent.brain.Q), total_reward, agent.brain.policy.epsilon.get(), goal)
            logger.info(optimal_summary + " * optimal * ")
            agent.set_eval(False)


            if self.check_stop_conditions(optimal_stats, eval=eval):
                break

            if self.cur_episode%100==0 and not eval:
                agent.save(self.agent_data_dir)
                self.save()

            self.cur_episode = ep

        if not eval:
            self.save()
            agent.save(self.agent_data_dir)

            stats.to_csv(self.data_dir + "/" + stats.name + "_" + str(time.time()) + "_" + str(random.random())[2:7])
            optimal_stats.to_csv(self.data_dir + "/" + optimal_stats.name + "_" + str(time.time()) + "_" + str(random.random())[2:7])

        return stats, optimal_stats

    def train_loop(self, render:bool=False):
        env = self.env
        agent = self.agent

        stop_condition = False
        info = {"goal": False}

        state = env.reset()
        action = agent.start(state)
        obs = None
        if render: env.render()

        while not stop_condition:
            state2, reward, done, info = env.step(action)
            if render: env.render()

            stop_condition = self.check_episode_stop_conditions(done, info.get("goal", False))
            obs = agent.observe(state, action, reward, state2, is_terminal_state=stop_condition)

            if stop_condition:
                break

            action = agent.step(obs)
            agent.update()
            state = state2

        agent.end(obs)
        steps, tot_reward = agent.brain.episode_iteration, agent.brain.total_reward
        return steps, tot_reward, self.is_goal(info)

    def check_stop_conditions(self, stats:StatsManager, *args, eval=False, **kwargs):
        return not eval and all([s.check_condition(stats_manager=stats) for s in self.stop_conditions])

    def check_episode_stop_conditions(self, done, goal, *args, **kwargs):
        return done or goal

    def is_goal(self, info, *args, **kwargs):
        return info.get("goal", False)

    def save(self):
        with open(self.data_dir + "/trainer.pkl", "wb") as fout:
            pickle.dump(self, fout)

    def reset(self):
        self.cur_episode = 0
        self.stats.reset()
        self.optimal_stats.reset()



import pickle
from abc import ABC

from typing import Tuple

from rltg.agents.brains.Brain import Brain
from rltg.agents.feature_extraction import RobotFeatureExtractor
from rltg.utils.misc import AgentObservation
from rltg.utils.RewardShaping import RewardShaping


class Agent(ABC):
    def __init__(self, sensors: RobotFeatureExtractor, brain:Brain, reward_shapers:Tuple[RewardShaping]=(), eval:bool=False):
        """
        :param sensors:             feature extractor from the environment observation.
        :param brain:               The Brain, a wrapper for the actual learning algorithm.
        :param eval:                If eval is True, prevent the brain to learn (testing mode).
                                    otherwise work as usual (learning mode)
                                    default: False

        :raises ValueError          if the space of the feature extractor output is different from the space of
                                    the brain input.
        """
        if brain.observation_space and sensors.output_space != brain.observation_space:
            raise ValueError("space dimensions are not compatible.")
        self.sensors = sensors
        self.brain = brain
        self.reward_shapers = reward_shapers
        self.eval = eval

    def set_eval(self, eval:bool):
        """Setter method for "eval" field."""
        self.eval = eval
        self.brain.set_eval(eval)

    def observe(self, state, action, world_reward, state2, **kwargs):
        """Called at each observation. """
        features_1 = self.sensors(state)
        features_2 = self.sensors(state2)
        reward = world_reward + sum(rs(features_1, features_2) for rs in self.reward_shapers)
        return self._observe(features_1, action, reward, features_2, **kwargs)

    def _observe(self, features, action, reward, features2, **kwargs):
        obs = AgentObservation(features, action, reward, features2)
        self.brain.observe(obs)
        return obs

    def update(self):
        """Called at the end of each iteration.
        It MUST be called only once for iteration."""
        self.brain.update()

    def start(self, state):
        features = self.sensors(state)
        return self.brain.start(features)

    def step(self, obs: AgentObservation):
        return self.brain.step(obs)

    def end(self, obs: AgentObservation):
        """Called at the end of each episode.
        It MUST be called only once for episode."""
        self.brain.end(obs)

    def save(self, filepath):
        with open(filepath + "/agent.pkl", "wb") as fout:
            pickle.dump(self, fout)

    @staticmethod
    def load(filepath):
        with open(filepath + "/agent.pkl", "rb") as fin:
            agent = pickle.load(fin)
        return agent

    def reset(self):
        self.brain.reset()



from collections import defaultdict

from rltg.agents.parameters.Parameter import Parameter
from rltg.utils.misc import mydefaultdict


class Eligibility(Parameter):

    def __init__(self, lambda_, gamma):
        super().__init__()
        self.lambda_ = lambda_
        self.gamma = gamma

        self.traces = mydefaultdict(0.0)

    def get(self, state, action):
        return self.traces[(state, action)]

    def to_zero(self, state, action):
        self.pop(state, action)

    def to_one(self, state, action):
        self.reset_state_action(state, action)

    def reset_state_action(self, state, action):
        self.traces[(state, action)] = 1

    def pop(self, state, action):
        self.traces.pop((state, action))

    def update(self, state, action, *args, **kwargs):
        self.traces[(state, action)] = self.gamma * self.lambda_ * self.traces[(state, action)]
        if self.traces[(state, action)] < 1e-4:
            self.pop(state, action)

    def reset(self):
        self.traces = {}



from copy import copy
from flloat.base.Alphabet import Alphabet
from flloat.base.Symbol import Symbol
from flloat.syntax.ldlf import LDLfFormula
from pythomata.base.DFA import DFA
from typing import Set

from rltg.logic.RewardAutomaton import RewardAutomaton
from rltg.utils.misc import compute_levels, _potential_function


class CompleteRewardAutomaton(DFA, RewardAutomaton):
    def __init__(self, dfa:DFA, alphabet:Alphabet, f:LDLfFormula, reward, gamma=0.99):
        super().__init__(
            dfa.alphabet,
            dfa.states,
            dfa.initial_state,
            dfa.accepting_states,
            dfa.transition_function
        )
        self._dfa = dfa
        self.alphabet = alphabet
        self.f = f
        self.reward = reward

        self.reachability_levels,   \
        self.max_level,             \
        self.failure_states = self._compute_levels()

        self.gamma = gamma

    @staticmethod
    def _fromFormula(alphabet:Set[Symbol], f:LDLfFormula, reward, gamma=0.99):
        dfa = f.to_automaton(alphabet, determinize=True, minimize=True)
        return CompleteRewardAutomaton(dfa, dfa.alphabet, f, reward, gamma=gamma)

    def get_formula_reward(self):
        return self.reward

    def complete(self):
        return CompleteRewardAutomaton(self._dfa.complete(), self.alphabet, self.f, self.reward, gamma=self.gamma)

    def minimize(self):
        return CompleteRewardAutomaton(self._dfa.minimize(), self.alphabet, self.f, self.reward, gamma=self.gamma)

    def trim(self):
        return CompleteRewardAutomaton(self._dfa.trim(), self.alphabet, self.f, self.reward, gamma=self.gamma)

    def potential_function(self, q, is_terminal_state=False):
        return _potential_function(q, self._dfa.initial_state, self.reachability_levels, self.reward,
                                   is_terminal_state=is_terminal_state)

    def _compute_levels(self):
        return compute_levels(self._dfa, self._dfa.accepting_states)
